""" Authorization based on Allowed HOSTS lists."""
import fnmatch
import logging
from .base import BaseAuthzHandler
from navigator.conf import ALLOWED_HOSTS
from aiohttp import web


class authz_allow_hosts(BaseAuthzHandler):
    """
    Allowed Hosts.
       Check if Origin is on the Allowed Hosts List.
    """

    async def check_authorization(self, request: web.Request) -> bool:
        origin = request.host if request.host else request.headers["origin"]
        for key in ALLOWED_HOSTS:
            if fnmatch.fnmatch(origin, key):
                logging.debug(
                    f'Authorization based on ALLOW HOST {key}'
                )
                return True
        return False



"""Django Session Backend.

Navigator Authentication using Anonymous Backend
"""
import logging
import asyncio
from aiohttp import web, hdrs
from platformdirs import user_cache_dir
from .base import BaseAuthBackend
import uuid
from navigator.conf import (
    CREDENTIALS_REQUIRED,
    AUTH_SESSION_OBJECT
)
from navigator.auth.sessions import get_session, new_session
from navigator.exceptions import (
    NavException,
    FailedAuth,
    InvalidAuth
)
from navigator.auth.models import AuthUser, guest


class AnonymousUser(AuthUser):
    first_name: str = 'Anonymous'
    last_name: str = 'User'


class NoAuth(BaseAuthBackend):
    """Basic Handler for No authentication."""

    user_attribute: str = "userid"

    def __init__(
        self,
        user_attribute: str = "userid",
        userid_attribute: str = "userid",
        password_attribute: str = "password",
        credentials_required: bool = False,
        authorization_backends: tuple = (),
        **kwargs,
    ):
        super(NoAuth, self).__init__(
            user_attribute,
            userid_attribute,
            password_attribute,
            credentials_required,
            authorization_backends,
            **kwargs
        )

    async def check_credentials(self, request):
        """ Authentication and create a session."""
        return True

    def get_userdata(self):
        key = uuid.uuid4().hex
        userdata = {
            AUTH_SESSION_OBJECT: {
                "session": key,
                self.user_property: key,
                self.username_attribute: "Anonymous",
                "first_name": "Anonymous",
                "last_name": "User"
            }
        }
        return [ userdata, key ]

    async def authenticate(self, request):
        userdata, key = self.get_userdata()
        user = AnonymousUser(data=userdata[AUTH_SESSION_OBJECT])
        user.id = key
        user.groups([guest])
        user.set(self.username_attribute, 'Anonymous')
        print(f'User Created: ', user)
        payload = {
            self.session_key_property: key,
            self.user_property: None,
            self.username_attribute: "Anonymous",
            **userdata
        }
        await self.remember(
            request, key, userdata, user
        )
        request.user = user
        token = self.create_jwt(data=payload)
        print('END AUTH')
        return {
            "token": token,
            self.session_key_property: key,
            self.username_attribute: "Anonymous",
            **userdata
        }

    async def auth_middleware(self, app, handler):
        """
         NoAuth Middleware.
         Description: Basic Authentication for NoAuth, Basic, Token and Django.
        """
        async def middleware(request):
            jwt_token = None
            try:
                authz = await self.authorization_backends(app, handler, request)
                if authz:
                    # Authorization Exception
                    return await authz
            except Exception as err:
                logging.exception(
                    f'Error Processing Authorization Middlewares {err!s}'
                )
            print('START MIDDLEWARE')
            try:
                auth = request.get('authenticated', False)
                if auth is True:
                    # already authenticated
                    return await handler(request)
            except KeyError:
                pass
            try:
                print('START DECODING')
                tenant, payload = self.decode_token(request)
                print('PAYLOAD ', payload)
                if payload:
                    # load session information
                    session = await get_session(request, payload, new = False)
                    print('SESSION ', session)
                    try:
                        request.user = session.decode('user')
                        print('USER> ', request.user, type(request.user))
                        request.user.is_authenticated = True
                        request['authenticated'] = True
                    except Exception:
                        logging.error('Missing User Object from Session')
            except NavException as err:
                logging.error('Auth Middleware: Invalid Signature or secret')
                pass # NoAuth can pass silently when no token was generated
            except Exception as err:
                logging.error(f"Bad Request: {err!s}")
                pass
            print('END MIDDLEWARE')
            return await handler(request)

        return middleware



"""NAVIGATOR.

Navigator is a simple framework to build asyncio-based applications, full
of features similar to django as Applications, domains and sub-domains.

Run:
    Run Navigator works simply to load run.py::

        $ python run.py

Section breaks are created by resuming unindented text. Section breaks
are also implicitly created anytime a new section starts.

Attributes:
    * WIP

TODO:
    * Work with asgi loaders
    * You have to also use ``sphinx.ext.todo`` extension

.. More information in:
https://github.com/phenobarbital/navigator-api

"""
from .navigator import Application, Response
from .version import (
    __title__, __description__, __version__, __author__
)

__all__ = ("Application", "Response", )



"""
RNC Token Authorization Middleware.
This token use a RNC algorithm to create a token-based authorization
for Navigator.
"""
from aiohttp import web
from typing import (
    Optional,
    Coroutine,
    Tuple
)
from navigator.libs.cypher import Cipher
from navigator.conf import (
    PARTNER_KEY,
    CYPHER_TYPE,
    PARTNER_SESSION_TIMEOUT
)
from .abstract import base_middleware

# TODO: add expiration logic when read the token
CIPHER = Cipher(PARTNER_KEY, type=CYPHER_TYPE)


class troctoken_middleware(base_middleware):
    def __init__(
        self,
        user_fn: Optional[Coroutine] = None,
        protected_routes: Optional[Tuple] = tuple()
    ):
        """
        Check if an Auth Token was provided and returns based on
        an user Callback Function.

        The Optional Callback can receive the Payload of the deciphered token
        and the Request.
        """
        if user_fn is not None and not callable(user_fn):
            raise RuntimeError(
                f"If defined, User Function {user_fn!s} need to be Callable."
            )
        self._fn = user_fn
        if protected_routes:
            self.protected_routes = protected_routes

    async def middleware(self, app, handler):
        @web.middleware
        async def middleware(request):
            if await self.valid_routes(request):
                return await handler(request)
            try:
                token, scheme = self.get_authorization_header(request)
            except Exception as err:
                print(err)
                token = None
            if self.path_protected(request): # is a protected site.
                if not token:
                    raise web.HTTPForbidden(
                        reason='Invalid authorization Token',
                    )
                try:
                    payload = CIPHER.decode(
                        passphrase=token
                    )
                    if not payload:
                        raise web.HTTPForbidden(
                            reason="Invalid authorization Token"
                        )
                except ValueError as err:
                    raise web.HTTPUnauthorized(
                        reason="Token Decryption Error"
                    )
                except Exception as err:
                    raise web.HTTPBadRequest(
                        reason=f"Token Decryption Error: {err}"
                    )
                try:
                    user = await self._fn(payload, request)
                    if user:
                        request[self.user_property] = user
                        request.user = user
                        print(user)
                    else:
                        raise web.HTTPForbidden(
                            reason='Access Restricted'
                        )
                except Exception as err:
                    raise web.HTTPBadRequest(
                        reason=f'Exception on Callable called by {__name__!s} {err!s}'
                    )
            return await handler(request)
        return middleware



import base64
import rapidjson
import logging
import time
from aiohttp import web
from typing import (
    Optional,
    Coroutine,
    Tuple
)
from navigator.conf import (
    SESSION_TIMEOUT,
    SESSION_KEY,
    SECRET_KEY,
    SESSION_PREFIX,
    CREDENTIALS_REQUIRED,
    SESSION_STORAGE,
    SESSION_USER_PROPERTY
)
from .abstract import base_middleware


class django_middleware(base_middleware):
    def __init__(
        self,
        user_fn: Optional[Coroutine] = None,
        protected_routes: Optional[Tuple] = tuple()
    ):
        """
        Extract an user Session from Django using a Middleware.

        The Optional Callback can receive the Payload of the deciphered token
        and the Request.
        """
        if user_fn is not None and not callable(user_fn):
            raise RuntimeError(
                f"If defined, User Function {user_fn!s} need to be Callable."
            )
        self._fn = user_fn
        if protected_routes:
            self.protected_routes = protected_routes

    def get_authorization_header(self, request: web.Request):
        sessionid = request.headers.get("x-sessionid", None)
        if not sessionid:
            raise web.HTTPBadRequest(
                reason=f'Django Middleware: use Header different from X-Sessionid is not available'
            )
        return sessionid

    async def middleware(self, app, handler):
        @web.middleware
        async def middleware(request):
            if await self.valid_routes(request):
                return await handler(request)
            try:
                sessionid = self.get_authorization_header(request)
            except Exception as err:
                sessionid = None
            if self.path_protected(request): # is a protected site.
                if not sessionid:
                    raise web.HTTPForbidden(
                        reason='Django Middleware: Invalid authorization Token',
                    )
                try:
                    redis = app["redis"]
                    payload = await redis.get(f"{SESSION_PREFIX}:{sessionid}")
                    if not payload:
                        raise web.HTTPBadRequest(
                            reason="Django Middleware: Invalid Django Session"
                        )
                    data = base64.b64decode(payload)
                    session_data = data.decode("utf-8").split(":", 1)
                    user = rapidjson.loads(session_data[1])
                    data = {
                        "key": sessionid,
                        "session_id": session_data[0],
                        **user
                    }
                    try:
                        user = await self._fn(data, request)
                        if user:
                            request[self.user_property] = user
                            request.user = user
                        else:
                            raise web.HTTPForbidden(
                                reason='Access Restricted'
                            )
                    except Exception as err:
                        raise web.HTTPBadRequest(
                            reason=f'Exception on Callable called by {__name__!s} {err!s}'
                        )
                except Exception as err:
                    raise web.HTTPBadRequest(
                        reason=f'Django Middleware: Error decoding: {err!s}'
                    )
            return await handler(request)
        return middleware



from abc import abstractmethod
from textwrap import dedent

from aiohttp import web


class AbstractHandler(object):
    _template = "<!doctype html><head></head><body></body>"

    def __init__(self):
        self._template = dedent(self._template)

    async def index(self, request):
        response = web.Response(body="")
        return response

    def html_response(text: str):
        return web.Response(text=text, content_type="text/html")

    @abstractmethod
    async def configure(self, app):
        pass



from pathlib import Path
from io import BytesIO
from jinja2 import (
    Environment,
    FileSystemLoader,
    ModuleLoader
)
from typing import (
    Dict,
    Optional
)

# TODO: implementing a bytecode-cache on redis or memcached
jinja_config = {
    'autoescape': True,
    'enable_async': False,
    'extensions': [
        'jinja2.ext.autoescape',
        'jinja2.ext.with_',
        'jinja2.ext.i18n'
    ]
}

class TemplateParser(object):
    """
    TemplateParser.

    This is a wrapper for the Jinja2 template engine.
    """
    path = None
    template = None
    filename = None
    config = None
    env = None
    cache = None
    def __init__(self, directory: Path, **kwargs):
        self.path = directory.resolve()
        if not self.path.exists():
            raise RuntimeError(
                f'NAVIGATOR: template directory {self.path} does not exist'
            )
        if 'config' in kwargs:
            self.config = {**jinja_config, **kwargs['config']}
        else:
            self.config = jinja_config
        # create loader:
        templateLoader = FileSystemLoader(
            searchpath=[str(self.path)]
        )
        # initialize the environment
        try:
            # TODO: check the bug ,encoding='ANSI'
            self.env = Environment(
                loader=templateLoader,
                **self.config
            )
            compiled_path = str(self.path.joinpath('.compiled'))
            self.env.compile_templates(
                target=compiled_path, zip='deflated'
            )
        except Exception as err:
            raise RuntimeError(
                f'NAV: Error loading Template Environment: {err}'
            )

    def get_template(self, filename: str):
        """
        Get a template from Template Environment using the Filename.
        """
        self.template = self.env.get_template(str(filename))
        return self.template

    @property
    def environment(self):
        """
        Property to return the current Template Environment.
        """
        return self.env

    def render(self, filename: str, params: Optional[Dict] = {}) -> str:
        result = None
        try:
            self.template = self.env.get_template(
                str(filename)
            )
            result = self.template.render(**params)
        except Exception as err:
            raise RuntimeError(
                f'NAV: Error rendering template: {filename}, error: {err}'
            )
        finally:
            return result



""" Abstract handler for Authorization Middleware."""

import asyncio
from aiohttp import web
from abc import ABC, abstractmethod


class BaseAuthzHandler(ABC):
    @abstractmethod
    async def check_authorization(self, request: web.Request) -> bool:
        pass



import base64
import codecs
import sys
from typing import Any, Callable, List, Literal

from Crypto import Random
from Crypto.Cipher import AES
from rncryptor import DecryptionError, RNCryptor

base64encoded = False


class Cipher(object):
    """Can Encode/Decode a string using AES-256 or RNCryptor."""

    key: str = ""
    type: str = "AES"
    iv: str = ""
    BS: int = 16
    cipher: Any = None

    def __init__(self, key: str, type: str = "AES"):
        self.key = key
        self.type = type
        if type == "AES":
            self.iv = Random.new().read(AES.block_size)
            self.cipher = AES.new(self.key, AES.MODE_CFB, self.iv)
        elif type == "RNC":
            self.cipher = RNCryptor()
        else:
            sys.exit("Not implemented")

    def encode(self, message: Any) -> str:
        if self.type == "AES":
            msg = self.iv + self.cipher.encrypt(message.encode("utf-8"))
        elif self.type == "RNC":
            msg = self.cipher.encrypt(message, self.key)
            if base64encoded:
                msg = base64.b64encode(msg)
        else:
            return ""
        if msg:
            return codecs.encode(msg, "hex").decode("utf-8")
        else:
            return ""

    def decode(self, passphrase: Any) -> bytes:
        msg = codecs.decode(passphrase, "hex")
        if self.type == "AES":
            try:
                return self.cipher.decrypt(msg)[len(self.iv) :].decode("utf-8")
            except Exception as e:
                print(e)
                raise (e)
        elif self.type == "RNC":
            try:
                msg = self.cipher.decrypt(msg, self.key)
                if base64encoded:
                    return base64.b64decode(msg, validate=True)
                else:
                    return msg
            except DecryptionError:
                raise ValueError("Error decoding message")
            except Exception as e:
                print(e)
                raise (e)
        else:
            return b""



"""Django Session Backend.

Navigator Authentication using API Token
description: Single API Token Authentication
"""
import base64
import rapidjson
import logging
import asyncio
import jwt
import uuid
from aiohttp import web, hdrs
from asyncdb import AsyncPool
from .base import BaseAuthBackend
from navigator.exceptions import NavException, UserDoesntExists, InvalidAuth
from datetime import datetime, timedelta
from navigator.conf import (
    SESSION_URL,
    SESSION_TIMEOUT,
    SECRET_KEY,
    PARTNER_KEY,
    JWT_ALGORITHM,
    SESSION_PREFIX,
    default_dsn,
    AUTH_SESSION_OBJECT,
    AUTH_TOKEN_ISSUER
)
from navigator.auth.sessions import get_session
from navigator.auth.models import AuthUser
from asyncdb.models import Column

class TokenUser(AuthUser):
    tenant: str
    programs: list = Column(default_factory=[])


class TokenAuth(BaseAuthBackend):
    """API Token Authentication Handler."""

    _pool = None

    def configure(self, app, router, handler):
        super(TokenAuth, self).configure(app, router, handler)

    async def on_startup(self, app: web.Application):
        try:
            kwargs = {
                "min_size": 1,
                "server_settings": {
                    "application_name": 'AUTH-NAV',
                    "client_min_messages": "notice",
                    "max_parallel_workers": "48",
                    "jit": "off",
                    "statement_timeout": "3600",
                    "effective_cache_size": "2147483647"
                },
            }
            self._pool = AsyncPool(
                "pg",
                dsn=default_dsn,
                **kwargs
            )
            await self._pool.connect()
        except Exception as err:
            print(err)
            raise Exception(
                f"Error Auth Token: please enable Connection Pool on AppHandler: {err}"
            )

    async def on_cleanup(self, app: web.Application):
        """
        Close the Pool when shutdown App.
        """
        try:
            await self._pool.close()
        except Exception as err:
            logging.exception(err)

    async def get_payload(self, request):
        token = None
        tenant = None
        id = None
        try:
            if "Authorization" in request.headers:
                try:
                    scheme, id = (
                        request.headers.get("Authorization").strip().split(" ", 1)
                    )
                except ValueError:
                    raise NavException(
                        "Invalid authorization Header",
                        state=400
                    )
                if scheme != self.scheme:
                    raise NavException(
                        "Invalid Authorization Scheme",
                        state=400
                    )
                try:
                    tenant, token = id.split(":")
                except ValueError:
                    token = id
        except Exception as e:
            print(e)
            return None
        return [tenant, token]

    async def reconnect(self):
        if not self.connection or not self.connection.is_connected():
            await self.connection.connection()

    async def authenticate(self, request):
        """ Authenticate, refresh or return the user credentials."""
        try:
            tenant, token = await self.get_payload(request)
            print(tenant, token)
            logging.debug(f"Tenant ID: {tenant}")
        except Exception as err:
            raise NavException(err, state=400)
        if not token:
            raise InvalidAuth("Invalid Credentials", state=401)
        else:
            payload = jwt.decode(
                token, PARTNER_KEY, algorithms=[JWT_ALGORITHM], leeway=30
            )
            logging.debug(f"Decoded Token: {payload!s}")
            data = await self.check_token_info(request, tenant, payload)
            if not data:
                raise InvalidAuth(f"Invalid Session: {token!s}", state=401)
            # getting user information
            # making validation
            try:
                u = data["name"]
                username = data["partner"]
                grants = data["grants"]
                programs = data["programs"]
            except KeyError as err:
                print(err)
                raise InvalidAuth(
                    f"Missing attributes for Partner Token: {err!s}", state=401
                )
            # TODO: Validate that partner (tenants table):
            try:
                userdata = dict(data)
                id = data["name"]
                user = {
                    "name": data["name"],
                    "partner": username,
                    "issuer": AUTH_TOKEN_ISSUER,
                    "programs": programs,
                    "grants": grants,
                    "tenant": tenant,
                    "id": data["name"],
                    "user_id": id,
                }
                userdata[self.session_key_property] = id
                usr = TokenUser(data=userdata)
                usr.id = id
                usr.set(self.username_attribute, id)
                usr.programs = programs
                usr.tenant = tenant
                print(f'User Created: ', usr)
                # saving user-data into request:
                await self.remember(
                    request, id, userdata, usr
                )
                token = self.create_jwt(data=user)
                return {
                    "token": f"{tenant}:{token}",
                    **user
                }
            except Exception as err:
                print(err)
                return False

    async def check_credentials(self, request):
        pass

    async def check_token_info(self, request, tenant, payload):
        try:
            name = payload["name"]
            partner = payload["partner"]
        except KeyError as err:
            return False
        sql = f"""
        SELECT name, partner, grants, programs FROM troc.api_keys
        WHERE name='{name}' AND partner='{partner}'
        AND enabled = TRUE AND revoked = FALSE AND '{tenant}'= ANY(programs)
        """
        try:
            result = None
            async with await self._pool.acquire() as conn:
                result, error = await conn.queryrow(sql)
            if error or not result:
                return False
            else:
                return result
        except Exception as err:
            logging.exception(err)
            return False

    async def auth_middleware(self, app, handler):
        async def middleware(request):
            request.user = None
            authz = await self.authorization_backends(app, handler, request)
            if authz:
                return await authz
            print('START MIDDLEWARE')
            try:
                auth = request.get('authenticated', False)
                if auth is True:
                    # already authenticated
                    return await handler(request)
            except KeyError:
                pass
            tenant, jwt_token = await self.get_payload(request)
            if jwt_token:
                try:
                    payload = jwt.decode(
                        jwt_token, PARTNER_KEY, algorithms=[JWT_ALGORITHM], leeway=30
                    )
                    logging.debug(f"Decoded Token: {payload!s}")
                    result = await self.check_token_info(request, tenant, payload)
                    if not result:
                        raise web.HTTPForbidden(
                            reason="API Key Not Authorized",
                        )
                    else:
                        # TRUE because if data doesnt exists, returned
                        session = await get_session(request, payload, new = True)
                        session["grants"] = result["grants"]
                        session["partner"] = result["partner"]
                        session["tenant"] = tenant
                        request.user = session.decode('user')
                        print('USER> ', request.user, type(request.user))
                        request.user.is_authenticated = True
                        request['authenticated'] = True
                except (jwt.DecodeError, jwt.InvalidTokenError) as err:
                    logging.exception(f"Invalid authorization token: {err!r}")
                except (jwt.ExpiredSignatureError) as err:
                    logging.exception(f"TokenAuth: token expired: {err!s}")
                except Exception as err:
                    print(err, err.__class__.__name__)
            return await handler(request)

        return middleware



"""
Model System for Navigator Auth.

Model for User, Group and Roles for Navigator Auth.
"""
from email.policy import default
import re
import uuid
import asyncio
import logging
import slugify
from asyncdb.models import Model, Column, Field
from typing import (
    Optional,
    List,
    Dict,
    Union,
    Tuple,
    Any,
    Callable
)
from dataclasses import InitVar
from datetime import datetime
from navigator.conf import default_dsn, USERS_TABLE


class User(Model):
    """Basic User notation."""

    user_id: int = Column(required=False, primary_key=True)
    first_name: str
    last_name: str
    email: str = Column(required=False, max=254)
    password: str = Column(required=False, max=128)
    last_login: datetime = Column(required=False)
    username: str = Column(required=False)
    is_superuser: bool = Column(required=True, default=False)
    is_active: bool = Column(required=True, default=True)
    is_new: bool = Column(required=True, default=True)
    title: str = Column(equired=False, max=90)
    registration_key: str = Column(equired=False, max=512)
    reset_pwd_key: str = Column(equired=False, max=512)
    avatar: str = Column(max=512)

    def __getitem__(self, item):
        return getattr(self, item)

    @property
    def display_name(self):
        return f"{self.first_name} {self.last_name}"

    class Meta:
        driver = "pg"
        dsn: str = default_dsn
        name = USERS_TABLE
        schema = "public"
        strict = True
        frozen = False
        connection = None


# TODO: add autoincrement feature, read the last id and plus 1
class Group(Model):
    id: int = Column(required=True, primary_key=True)
    name: int = Column(required=True)

    class Meta:
        driver = "pg"
        dsn: str = default_dsn
        name = "auth_group"
        schema = "public"
        strict = True
        frozen = False
        connection = None


# class UserGroup(Model):
#     member_id: int = Column(required=True, primary_key=True)
#     user_id: User = Column(required=True)
#     group_id: Group = Column(required=True)
#
#     class Meta:
#         driver = 'pg'
#         name = 'auth_membership'
#         schema = 'public'
#         app_label = 'troc'
#         strict = True
#         frozen = False
#
# class Role(Model):
#     role_id: int = Column(required=True, primary_key=True)
#     rolename: str = Column(required=True, unique=True)
#     role: str = Column(required=True, unique=True, comment="Role Slug")
#     program_id: int = Column(required=False, default=1)
#
#     class Meta:
#         driver = 'pg'
#         name = 'auth_roles'
#         schema = 'public'
#         app_label = 'troc'
#         strict = True
#         frozen = False

class Organization(Model):
    org_id: str
    org_name: str
    slug: str
    
    def __post_init__(self) -> None:
        super(Organization, self).__post_init__()
        if not self.slug:
            self.slug = slugify(self.org_name)
    class Meta:
        strict = False
        frozen = False
        connection = None
        

class Program(Model):
    program_id: int
    program_name: str = Column(required=True)
    program_slug: str = Column(default='')
    
    def __post_init__(self) -> None:
        super(Program, self).__post_init__()
        if not self.program_slug:
            self.program_slug = slugify(self.program_name)

    class Meta:
        strict = False
        frozen = False
        connection = None


class AuthGroup(Model):
    """AuthGroup.
    
    Association (group) were users belongs to.
    """
    group_id: int
    group_name: str = Column(required=True)

    class Meta:
        strict = False
        frozen = False
        connection = None


guest = AuthGroup(group_id = 0, group_name = 'guest')

class AuthUser(Model):
    """AuthUser

    Model for any Authenticated User.
    """
    id: Any
    first_name: str
    last_name: str
    email: str
    username: str
    # group: List[AuthGroup]
    # organizations: List[Organization]
    # programs: List[Program]
    enabled: bool = Column(required=True, default=True)
    superuser: bool = Column(required=True, default=False)
    data: InitVar[Dict] = Column(required=False, default_factory=dict)
    userdata: Dict  = Column(required=False, default_factory={})
    is_authenticated: bool = Column(equired=False, default=False)

    class Meta:
        strict = False
        frozen = False
        connection = None
    
    def __post_init__(self, data):
        self.userdata = data
        for key, value in data.items():
            self.create_field(key, value)

    def create_field(self, name: str, value: Any) -> None:
        # create a new Field on Model (when strict is False).            
        f = Field(required=False, default=value)
        f.name = name
        f.type = type(value)
        self.__columns__[name] = f
        setattr(self, name, value)
        
    def set(self, name: str, value: Any) -> None:
        # alias for "create_field"
        self.create_field(name, value)

    """
    User Methods.
    """
    def groups(self, grp: List):
        self.group = grp
        
    def organizations(self, orgs: List):
        self.orgs = orgs


import os
import sys
from datetime import datetime
from pathlib import Path
from navigator.conf import BASE_DIR
import json


class SafeDict(dict):
    def __missing__(self, key):
        return "{" + key + "}"


def load_schema(schema_name: str, directory: Path = ""):
    if not directory:
        directory = BASE_DIR.joinpath("models")
    model = Path.joinpath(directory, "{}.json".format(schema_name))
    if model.is_file():
        with open(model, mode="r") as jsonfile:
            try:
                return json.load(jsonfile)
            except Exception as e:
                raise TypeError("Empty Schema File")



import os
import sys
import asyncio
from typing import List
import json
import rapidjson
import inspect
from functools import partial
from aiohttp import web
from asyncdb.utils.encoders import BaseEncoder


async def run_cmd(cmd: List) -> str:
    print(f'Executing: {" ".join(cmd)}')
    process = await asyncio.create_subprocess_exec(*cmd, stdout=asyncio.subprocess.PIPE)
    out, error = await process.communicate()
    if error:
        raise Exception(error)
    return out.decode("utf8")


def escapeString(value):
    v = value if value != "None" else ""
    v = str(v).replace("'", "''")
    v = "'{}'".format(v) if type(v) == str else v
    return v


def json_response(response={}, headers={}, state=200, cls=None):
    if cls is not None:
        if inspect.isclass(cls):
            # its a class-based Encoder
            jsonfn = partial(json.dumps, cls=cls)
        else:
            # its a function
            jsonfn = partial(json.dumps, default=cls)
    else:
        jsonfn = partial(json.dumps, cls=BaseEncoder)
    obj = web.json_response(response, status=state, dumps=jsonfn)
    for header, value in headers.items():
        obj.headers[header] = value
    return obj


### printing functions
class colors:
    """
    Colors class.

       reset all colors with colors.reset;
       Use as colors.subclass.colorname.
    i.e. colors.fg.red or colors.fg.greenalso, the generic bold, disable,
    underline, reverse, strike through,
    and invisible work with the main class i.e. colors.bold
    """

    reset = "\033[0m"
    bold = "\033[01m"
    disable = "\033[02m"
    underline = "\033[04m"
    reverse = "\033[07m"
    strikethrough = "\033[09m"
    invisible = "\033[08m"

    class fg:
        """
        colors.fg.

        Foreground Color subClass
        """

        black = "\033[30m"
        red = "\033[31m"
        green = "\033[32m"
        orange = "\033[33m"
        blue = "\033[34m"
        purple = "\033[35m"
        cyan = "\033[36m"
        lightgrey = "\033[37m"
        darkgrey = "\033[90m"
        lightred = "\033[91m"
        lightgreen = "\033[92m"
        yellow = "\033[93m"
        lightblue = "\033[94m"
        pink = "\033[95m"
        lightcyan = "\033[96m"


def cPrint(msg, color=None, level="INFO"):
    try:
        if color is not None:
            coloring = colors.bold + getattr(colors.fg, color)
        elif level:
            if level == "INFO":
                coloring = colors.bold + colors.fg.green
            elif level == "SUCCESS":
                coloring = colors.bold + colors.fg.lightgreen
            elif level == "NOTICE":
                coloring = colors.fg.blue
            elif level == "DEBUG":
                coloring = colors.fg.lightblue
            elif level == "WARN":
                coloring = colors.bold + colors.fg.yellow
            elif level == "ERROR":
                coloring = colors.fg.lightred
            elif level == "CRITICAL":
                coloring = colors.bold + colors.fg.red
        else:
            coloring = colors.reset
    except Exception as err:
        print("Wrong color schema {}, error: {}".format(color, str(err)))
        coloring = colors.reset
    print(coloring + msg, colors.reset)



"""GoogleAuth.

Description: Backend Authentication/Authorization using Google AUTH API.
"""
import base64
import rapidjson
import logging
import asyncio
import jwt
from aiohttp import web, hdrs
from .base import BaseAuthBackend
from typing import List, Dict, Any
from navigator.exceptions import (
    NavException,
    UserDoesntExists,
    InvalidAuth
)
from aiogoogle import Aiogoogle
from aiogoogle.auth.utils import create_secret
from navigator.conf import (
    GOOGLE_CLIENT_ID,
    GOOGLE_CLIENT_SECRET,
    GOOGLE_API_SCOPES,
    AUTH_SESSION_OBJECT
)

class GoogleAuth(BaseAuthBackend):
    """GoogleAuth.

    Authentication Backend using Google+aiogoogle.
    """
    user_attribute: str = "user"
    username_attribute: str = "username"
    pwd_atrribute: str = "password"
    _credentials: Dict = {}
    google: Any = None

    def configure(self, app, router):
        # TODO: build the callback URL and append to routes
        self._credentials = {
            "client_id": GOOGLE_CLIENT_ID,
            "client_secret": GOOGLE_CLIENT_SECRET,
            "scopes": GOOGLE_API_SCOPES,
            "redirect_uri": "http://localhost:5000/auth/complete/google-oauth2/",
        }
        self.google = Aiogoogle(
            client_creds=self._credentials
        )
        # add the callback url
        router = app.router
        router.add_route(
            "GET",
            "/auth/complete/google-oauth2/",
            self.finish_auth,
            name="google_complete_login"
        )
        router.add_route(
            "GET",
            "/api/v1/auth/google/",
            self.authenticate,
            name="google_api_login"
        )
        super(GoogleAuth, self).configure(app, router)

    async def get_payload(self, request):
        ctype = request.content_type
        if request.method == "GET":
            try:
                user = request.query.get(self.username_attribute, None)
                password = request.query.get(self.pwd_atrribute, None)
                return [user, password]
            except Exception:
                return None
        elif ctype in ("multipart/mixed", "application/x-www-form-urlencoded"):
            data = await request.post()
            if len(data) > 0:
                user = data.get(self.username_attribute, None)
                password = data.get(self.pwd_atrribute, None)
                return [user, password]
            else:
                return None
        elif ctype == "application/json":
            try:
                data = await request.json()
                user = data[self.username_attribute]
                password = data[self.pwd_atrribute]
                return [user, password]
            except Exception:
                return None
        else:
            return None

    async def authenticate(self, request):
        """ Authenticate, refresh or return the user credentials."""
        user = None
        try:
            user, pwd = await self.get_payload(request)
            print('USER', user, pwd)
        except Exception as err:
            print(err)
            # raise NavException(err, state=400)
            # making validation
        self._state = (
            create_secret()
        )  # Shouldn't be a global or a hardcoded variable. should be tied to a session or a user and shouldn't be used more than once
        self._nonce = (
            create_secret()
        )  # Shouldn't be a global or a hardcoded variable. should be tied to a session or a user and shouldn't be used more than once
        if self.google.openid_connect.is_ready(self._credentials):
            uri = self.google.openid_connect.authorization_url(
                client_creds=self._credentials,
                state=self._state,
                nonce=self._nonce,
                access_type="offline",
                include_granted_scopes=True,
                login_hint=user,
                prompt="select_account",
            )
            # Step A: redirect
            logging.debug(f'Google URI: {uri!s}')
            return web.HTTPFound(uri)
        else:
            raise NavException(
                "Client doesn't have info for Authentication"
            )

    async def finish_auth(self, request):
        try:
            if request.query.get('error'):
                error = {
                    "error": request.query.get("error"),
                    "error_description": request.query.get("error_description"),
                }
                print("Error HERE: ", err, err.state)
                response = {
                    "message": "Google Login Error",
                    **error
                }
                return web.json_response(response, status=403)
        except Exception as err:
            response = {
                "message": "Google Login Error"
            }
            return web.json_response(response, status=501)
        if code:=request.query.get("code"):
            state = request.query.get("state")
            if state != self._state:
                response = {
                    "message": "Something wrong with Authentication State",
                    "error": "Authenticate Error"
                }
                return web.json_response(response, status=403)
            user_creds = await self.google.openid_connect.build_user_creds(
                grant=code,
                client_creds=self._credentials,
                nonce=self._nonce,
                verify=False,
            )
            # print(user_creds)
            userdata = await self.google.openid_connect.get_user_info(
                user_creds
            )
            print(userdata)
            id = userdata['id']
            payload = {
                "user_id": id,
                **userdata
            }
            userdata[self.session_key_property] = id
            # saving Auth data.
            await self.remember(
                request, id, userdata
            )
            # Create the User session.
            token = self.create_jwt(data=payload)
            data = {
                "token": user_creds['id_token_jwt'],
                **userdata
            }
            # TODO: saving Auth token and user_creds on a Database
            return web.json_response(data, status=200)
        else:
            response = {
                "message": "Something wrong with Google Callback",
                "error": "Authenticate Error"
            }
            return web.json_response(response, status=403)

    async def check_credentials(self, request):
        """ Authentication and create a session."""
        return True



"""Navigator Exceptions."""


class NavException(Exception):
    """Base class for other exceptions"""

    state: int = 0

    def __init__(self, message: str = '', state: int = 0, *args, **kwargs):
        super().__init__(message)
        self.message = message
        self.state = int(state)

    def __str__(self):
        return f"{self.message}"

    def get(self):
        return self.message


class UserDoesntExists(NavException):
    state: int = 404


class InvalidAuth(NavException):
    state: int = 401

class FailedAuth(NavException):
    state: int = 403


class ValidationError(NavException):
    state: int = 406


"""JWT Backend.

Navigator Authentication using JSON Web Tokens.
"""
import jwt
import hashlib
import base64
import secrets
from aiohttp import web
from .base import BaseAuthBackend
from datetime import datetime, timedelta
from navigator.exceptions import (
    NavException,
    FailedAuth,
    UserDoesntExists,
    InvalidAuth,
    ValidationError,
)
from navigator.conf import (
    SESSION_TIMEOUT,
    SECRET_KEY,
    AUTH_PWD_DIGEST,
    AUTH_PWD_ALGORITHM,
    AUTH_PWD_LENGTH,
    AUTH_PWD_SALT_LENGTH,
    AUTH_SESSION_OBJECT
)
from navigator.auth.models import AuthUser


class BasicUser(AuthUser):
    """BasicAuth.
    
    Basic authenticated user (user:password).
    """
    pass


# "%s$%d$%s$%s" % (algorithm, iterations, salt, hash)


class BasicAuth(BaseAuthBackend):
    """Basic User/password Authentication."""

    user_attribute: str = "user"
    pwd_atrribute: str = "password"

    async def validate_user(self, login: str = None, password: str = None):
        # get the user based on Model
        try:
            search = {self.username_attribute: login}
            user = await self.get_user(**search)
        except UserDoesntExists as err:
            raise UserDoesntExists(
                f"User {login} doesn't exists"
            )
        except Exception:
            raise
        try:
            # later, check the password
            pwd = user[self.pwd_atrribute]
        except KeyError:
            raise ValidationError(
                'Missing Password attribute on User Account'
            )
        try:
            if self.check_password(pwd, password):
                # return the user Object
                return user
            else:
                raise FailedAuth(
                    "Basic Auth: Invalid Credentials"
                )
        except Exception:
            raise
        return None

    def set_password(
        self,
        password: str,
        token_num: int = 6,
        iterations: int = 80000,
        salt: str = None
    ):
        if not salt:
            salt = secrets.token_hex(token_num)
        key = hashlib.pbkdf2_hmac(
            AUTH_PWD_DIGEST,
            password.encode("utf-8"),
            salt.encode("utf-8"),
            iterations,
            dklen=AUTH_PWD_LENGTH,
        )
        hash = base64.b64encode(key).decode("utf-8").strip()
        return f"{AUTH_PWD_ALGORITHM}${iterations}${salt}${hash}"

    def check_password(self, current_password, password):
        try:
            algorithm, iterations, salt, hash = current_password.split("$", 3)
        except ValueError:
            raise InvalidAuth('Basic Auth: Invalid Password Algorithm')
        assert algorithm == AUTH_PWD_ALGORITHM
        compare_hash = self.set_password(
            password,
            iterations=int(iterations),
            salt=salt,
            token_num=AUTH_PWD_SALT_LENGTH
        )
        return secrets.compare_digest(current_password, compare_hash)

    async def get_payload(self, request):
        ctype = request.content_type
        if request.method == "GET":
            try:
                user = request.query.get(self.username_attribute, None)
                password = request.query.get(self.pwd_atrribute, None)
                return [user, password]
            except Exception:
                return [None, None]
        elif ctype in ("multipart/mixed", "multipart/form-data", "application/x-www-form-urlencoded"):
            data = await request.post()
            if len(data) > 0:
                user = data.get(self.username_attribute, None)
                password = data.get(self.pwd_atrribute, None)
                return [user, password]
            else:
                return [None, None]
        elif ctype == "application/json":
            try:
                data = await request.json()
                user = data[self.username_attribute]
                password = data[self.pwd_atrribute]
                return [user, password]
            except Exception:
                return [None, None]
        else:
            return [None, None]

    async def authenticate(self, request):
        """ Authenticate, refresh or return the user credentials."""
        try:
            user, pwd = await self.get_payload(request)
        except Exception as err:
            raise NavException(err, state=400)
        if not pwd and not user:
            raise InvalidAuth(
                "Basic Auth: Invalid Credentials",
                state=401
            )
        else:
            # making validation
            try:
                user = await self.validate_user(login=user, password=pwd)
            except FailedAuth as err:
                raise FailedAuth(err)
            except UserDoesntExists as err:
                raise UserDoesntExists(err)
            except (ValidationError, InvalidAuth) as err:
                raise InvalidAuth(err, state=401)
            except Exception as err:
                raise NavException(
                    err, state=500
                )
            try:
                userdata = self.get_userdata(user)
                username = user[self.username_attribute]
                id = user[self.userid_attribute]
                userdata[self.username_attribute] = username
                userdata[self.session_key_property] = username
                usr = BasicUser(data=userdata[AUTH_SESSION_OBJECT])
                usr.id = id
                usr.set(self.username_attribute, username)
                print(f'User Created: ', usr)
                payload = {
                    self.user_property: user[self.userid_attribute],
                    self.username_attribute: username,
                    "user_id": id,
                    self.session_key_property: username
                }
                await self.remember(
                    request, username, userdata, usr
                )
                # Create the User session and returned.
                token = self.create_jwt(data=payload)
                return {
                    "token": token,
                    **userdata
                }
            except Exception as err:
                return False

    async def check_credentials(self, request):
        """ Using for check the user credentials to the backend."""
        pass



"""OktaAuth.

Description: Backend Authentication/Authorization using Okta Service.
"""
import rapidjson
import logging
import asyncio
from aiohttp import web, hdrs
from .base import BaseAuthBackend
from typing import List, Dict, Any
import requests
from navigator.exceptions import (
    NavException,
    UserDoesntExists,
    InvalidAuth,
    FailedAuth
)
from navigator.conf import (
    OKTA_CLIENT_ID,
    OKTA_CLIENT_SECRET,
    OKTA_DOMAIN,
    OKTA_APP_NAME,
    AUTH_SESSION_OBJECT
)
from okta_jwt_verifier import JWTVerifier

async def is_token_valid(token, issuer, client_id):
    jwt_verifier = JWTVerifier(issuer, client_id, 'api://default')
    try:
        await jwt_verifier.verify_access_token(token)
        return True
    except Exception:
        return False

async def is_id_token_valid(token, issuer, client_id, nonce):
    jwt_verifier = JWTVerifier(issuer, client_id, 'api://default')
    try:
        await jwt_verifier.verify_id_token(token, nonce=nonce)
        return True
    except Exception:
        return False

class OktaAuth(BaseAuthBackend):
    """OktaAuth.

    Description: Authentication Backend using Third-party Okta Service.
    """
    user_attribute: str = "user"
    username_attribute: str = "username"
    pwd_atrribute: str = "password"
    _credentials: Dict = {}
    _okta: Any = None
    _uri: str = None
    nonce: Any = None

    def configure(self, app, router):
        self.base_uri = f"https://{OKTA_DOMAIN}/oauth2/default/v1/authorize"
        self.redirect_uri = "http://nav-api.dev.local:5000/auth/complete/okta/"
        # async def _setup_okta(app):
        #     pass
        # asyncio.get_event_loop().run_until_complete(_setup_okta(app))
        # add the callback url
        router = app.router
        router.add_route(
            "GET",
            "/auth/complete/okta/",
            self.finish_auth,
            name="okta_complete_login"
        )
        router.add_route(
            "GET",
            "/api/v1/auth/okta/",
            self.authenticate,
            name="okta_api_login"
        )
        router.add_route(
            "GET",
            "/api/v1/auth/okta/logout",
            self.logout,
            name="okta_api_logout"
        )
        router.add_route(
            "GET",
            "/auth/okta/logout",
            self.finish_logout,
            name="okta_complete_logout"
        )
        self._issuer = f"https://{OKTA_DOMAIN}/oauth2/default"
        self._token_uri = f"https://{OKTA_DOMAIN}/oauth2/default/v1/token"
        self._introspection_uri = f"https://{OKTA_DOMAIN}/oauth2/default/v1/introspect"
        self._userinfo_uri = f"https://{OKTA_DOMAIN}/oauth2/default/v1/userinfo"
        # executing parent configurations
        super(OktaAuth, self).configure(app, router)

    async def get_payload(self, request):
        ctype = request.content_type
        if request.method == "GET":
            try:
                user = request.query.get(self.username_attribute, None)
                password = request.query.get(self.pwd_atrribute, None)
                return [user, password]
            except Exception:
                return None
        elif ctype in ("multipart/mixed", "application/x-www-form-urlencoded"):
            data = await request.post()
            if len(data) > 0:
                user = data.get(self.username_attribute, None)
                password = data.get(self.pwd_atrribute, None)
                return [user, password]
            else:
                return None
        elif ctype == "application/json":
            try:
                data = await request.json()
                user = data[self.username_attribute]
                password = data[self.pwd_atrribute]
                return [user, password]
            except Exception:
                return None
        else:
            return None

    async def authenticate(self, request):
        """ Authenticate, refresh or return the user credentials."""
        user = None
        try:
            # Build the URL
            APP_STATE = 'ApplicationState'
            self.nonce = 'SampleNonce'
            query_params = {
              "client_id": f"{OKTA_CLIENT_ID}",
              # "client_secret": f"{OKTA_CLIENT_SECRET}",
              "redirect_uri": self.redirect_uri,
              'scope': "openid email profile",
              'state': APP_STATE,
              'nonce': self.nonce,
              'response_type': 'code',
              'response_mode': 'query',
            }
            uri = "{base_uri}?{query_params}".format(
                base_uri=self.base_uri,
                query_params=requests.compat.urlencode(query_params)
            )
            # Step A: redirect
            logging.debug(f'Okta URI: {uri}')
            return web.HTTPFound(uri)
        except Exception as err:
            print('HERE: ', err)
            raise NavException(
                f"Client doesn't have info for Okta Authentication: {err}"
            )

    async def finish_auth(self, request):
        headers = {'Content-Type': 'application/x-www-form-urlencoded'}
        code = request.query.get("code")
        if not code:
            response = {
                "message": "Okta Code not accessible"
            }
            return web.json_response(response, status=403)
        # B.- processing the code
        query_params = {
            'grant_type': 'authorization_code',
            'code': code,
            'redirect_uri': self.redirect_uri
        }
        query_params = requests.compat.urlencode(query_params)
        print(query_params)
        try:
            exchange = requests.post(
                self._token_uri,
                headers=headers,
                data=query_params,
                auth=(
                    OKTA_CLIENT_ID,
                    OKTA_CLIENT_SECRET
                ),
            ).json()
        except Exception as err:
            print('ERR> ', err)
            response = {
                "message": "Okta: Error Getting User Profile information"
            }
            return web.json_response(response, status=403)
        print(exchange)
        # Get tokens and validate
        if not exchange.get("token_type"):
            response = {
                "message": "Okta: Unsupported token type. Should be 'Bearer'."
            }
            return web.json_response(response, status=403)
        # user data
        access_token = exchange["access_token"]
        id_token = exchange["id_token"]

        if not await is_token_valid(access_token, self._issuer, OKTA_CLIENT_ID):
            response = {
                "message": "Okta: Access Token Invalid."
            }
            return web.json_response(response, status=403)

        if not await is_id_token_valid(id_token, self._issuer, OKTA_CLIENT_ID, self.nonce):
            response = {
                "message": "Okta: ID Token Invalid."
            }
            return web.json_response(response, status=403)

        # Authorization flow successful, get userinfo and login user
        userdata = requests.get(
            self._userinfo_uri,
            headers={'Authorization': f'Bearer {access_token}'}
        ).json()
        print(userdata)
        userdata['id'] = userdata["sub"]
        userdata[self.session_key_property] = userdata["sub"]
        userdata['access_token'] = access_token
        userdata['id_token'] = id_token
        # get user data
        # unique_id = userdata["sub"]
        # user_email = userdata["email"]
        # user_name = userdata["given_name"]
        # TODO: Optional: get User info from Nav
        payload = {
            "user_id": userdata["sub"],
            **userdata
        }
        # saving Auth data.
        await self.remember(
            request, userdata["sub"], userdata
        )
        # Create the User session.
        token = self.create_jwt(data=payload)
        data = {
            "token": token,
            "access_token": access_token,
            **userdata
        }
        # TODO: saving Auth token and user_creds on a Database
        return web.json_response(data, status=200)

    async def logout(self, request):
        pass

    async def finish_logout(self, request):
        pass

    async def check_credentials(self, request):
        """ Authentication and create a session."""
        return True



"""Navigator.

Authentication Backends.
"""
from .noauth import NoAuth
from .basic import BasicAuth
from .troc import TrocToken
from .django import DjangoAuth
from .token import TokenAuth
from .google import GoogleAuth
from .okta import OktaAuth
from .adfs import ADFSAuth

__all__ = [
    "NoAuth",
    "BasicAuth",
    "TrocToken",
    "DjangoAuth",
    "TokenAuth",
    "GoogleAuth",
    "OktaAuth",
    "ADFSAuth"
]



from collections.abc import MutableMapping
from typing import Any, Union, List, Dict, Tuple, Iterator, Iterable

class ClassDict(MutableMapping):
    def __init__(self, data: Union[Tuple, Dict] = (), default: Any = None):
        self._columns: list = []
        self.mapping = {}
        self.default = default
        self.update(data)
        self._columns = list(self.mapping.keys())
        print(self.mapping, self._columns)

    def update(self, items: Iterable):
        for key, value in items.items():
            self.mapping[key] = value

    def __missing__(self, key):
        return self.default

    def items(self) -> zip:  # type: ignore
        return zip(self._columns, self.mapping)

    @property
    def keys(self) -> List:
        return self._columns

    def set(self, key, value) -> None:
        self.mapping[key] = value
        if not key in self._columns:
            self._columns.append(key)

    """
     Section: Simple magic methods
    """
    def __len__(self) -> int:
        return len(self.mapping)

    def __str__(self) -> str:
        return f"<{type(self).__name__}({self.mapping})>"

    def __repr__(self) -> str:
        return f"<{type(self).__name__}({self.mapping})>"

    def __contains__(self, key: str) -> bool:
        return key in self._columns

    def __delitem__(self, key) -> None:
        value = self[key]
        del self.mapping[key]
        self.pop(value, None)

    def __setitem__(self, key, value):
        self.mapping[key] = value
        if not key in self._columns:
            self._columns.append(key)

    def __getitem__(self, key: Union[str, int]) -> Any:
        """
        Sequence-like operators
        """
        try:
            return self.mapping[key]
        except (KeyError, TypeError):
            return None

    def __getattr__(self, attr: str) -> Any:
        """
        Attributes for dict keys
        """
        try:
            return self.__getitem__(attr)
        except KeyError:
            raise KeyError(
                f"User Error: invalid field name {attr} on {self.mapping!r}"
            )
        except TypeError:
            raise TypeError(
                f"User Error: invalid attribute value on {self.mapping!r} for {attr}"
            )

    def __iter__(self) -> Iterator:
        for value in self.mapping:
            yield value



"""Using Redis for Saving Session Storage."""
import time
import uuid
import logging
import aioredis
import asyncio
import rapidjson
from aiohttp import web
from .abstract import AbstractStorage, SessionData
from functools import wraps, partial
from navigator.conf import (
    SESSION_URL,
    SESSION_KEY,
    SESSION_OBJECT,
    SESSION_STORAGE
)

class RedisStorage(AbstractStorage):
    """Test version of Session Handler."""
    _redis = None

    async def setup_redis(self, app: web.Application):
        redis = aioredis.ConnectionPool.from_url(
                SESSION_URL,
                decode_responses=True,
                encoding='utf-8'
        )
        # print('REDIS CACHE:  ', SESSION_URL, redis)
        async def close_redis(app):
            await redis.disconnect(inuse_connections = True)
        app.on_cleanup.append(close_redis)
        return redis

    def configure_session(self, app: web.Application) -> None:
        super(RedisStorage, self).configure_session(app)
        self._encoder = partial(
            rapidjson.dumps, datetime_mode=rapidjson.DM_ISO8601
        )
        self._decoder = rapidjson.loads
        async def _make_mredis():
            try:
                self._redis = await self.setup_redis(app)
            except Exception as err:
                print(err)
                return False
        return asyncio.get_event_loop().run_until_complete(_make_mredis())

    async def get_session(self, request: web.Request, userdata: dict = {}) -> SessionData:
        try:
            session = request.get(SESSION_OBJECT)
        except Exception as err:
            logging.debug(f'Error on get Session: {err!s}')
            session = None
        if session is None:
            storage = request.get(SESSION_STORAGE)
            if storage is None:
                raise RuntimeError(
                    "Missing Configuration for Session Middleware."
                )
            session = await self.load_session(request, userdata)
        request[SESSION_OBJECT] = session
        request["session"] = session
        return session

    async def invalidate(self, request: web.Request, session: SessionData) -> None:
        conn = aioredis.Redis(connection_pool=self._redis)
        if not session:
            data = None
            session_id = request.get(SESSION_KEY, None)
            if session_id:
                data = await conn.get(session_id)
            if data is None:
                # nothing to forgot
                return True
        try:
            # delete the ID of the session
            result = await conn.delete(session.identity)
            session.invalidate() # invalidate this session object
        except Exception as err:
            logging.error(err)
            return False
        return True

    async def load_session(
        self,
        request: web.Request,
        userdata: dict = {},
        new: bool = False
    ) -> SessionData:
        """
        Load Session.

        Load User session from backend storage, or create one if
        doesnt exists.

        ---
        new: if False, new session is not created.
        """
        # TODO: first: for security, check if cookie csrf_secure exists
        # if not, session is missed, expired, bad session, etc
        try:
            conn = aioredis.Redis(connection_pool=self._redis)
        except Exception as err:
            logging.exception(f'Error loading Redis Session: {err!s}')
        session_id = request.get(SESSION_KEY, None)
        if not session_id:
            session_id = userdata.get(SESSION_KEY, None) if userdata else None
            # TODO: getting from cookie
        if session_id is None and new is False:
            return False
        # we need to load session data from redis
        print(f':::::: LOAD SESSION {session_id} ::::: ')
        try:
            data = await conn.get(session_id)
        except Exception as err:
            logging.error(f'Error Getting Session data: {err!s}')
            data = None
        if data is None:
            if new is True:
                print(':::: CREATING A NEW SESSION :::: ')
                # create a new session if not exists:
                return await self.new_session(request, userdata)
            else:
                return False
        try:
            data = self._decoder(data)
            session = SessionData(
                db=conn,
                identity=session_id,
                data=data,
                new=False,
                max_age=self.max_age
            )
        except Exception as err:
            print(f'ERROR: ::::::::::: ', err)
            logging.debug(err)
            session = SessionData(
                db=conn,
                identity=None,
                data=None,
                new=True,
                max_age=self.max_age
            )
        request[SESSION_KEY] = session_id
        session[SESSION_KEY] = session_id
        request[SESSION_OBJECT] = session
        request["session"] = session
        return session

    async def save_session(self,
        request: web.Request,
        response: web.StreamResponse,
        session: SessionData
    ) -> None:
        """Save the whole session in the backend Storage."""
        session_id = session.identity
        if not session_id:
            session_id = session.get(SESSION_KEY, None)
        if not session_id:
            session_id = self.id_factory()
        if session.empty:
            data = {}
        data = self._encoder(session.session_data())
        max_age = session.max_age
        expire = max_age if max_age is not None else 0
        try:
            conn = aioredis.Redis(connection_pool=self._redis)
            result = await conn.set(
                session_id, data, self.max_age
            )
        except Exception as err:
            print('Error Saving Session: ', err)
            logging.exception(err)
            return False

    async def new_session(
        self,
        request: web.Request,
        data: dict = None
    ) -> SessionData:
        """Create a New Session Object for this User."""
        session_id = request.get(SESSION_KEY, None)
        if not session_id:
            try:
                session_id = data[SESSION_KEY]
            except KeyError:
                session_id = self.id_factory()
        print(f':::::: START CREATING A NEW SESSION {session_id} ::::: ')
        if not data:
            data = {}
        # saving this new session on DB
        try:
            conn = aioredis.Redis(connection_pool=self._redis)
            t = time.time()
            data['created'] = t
            data['last_visit'] = t
            data["last_visited"] = f"Last visited: {t!s}"
            data[SESSION_KEY] = session_id
            dt = self._encoder(data)
            result = await conn.set(
                session_id, dt, self.max_age
            )
            logging.info(f'Creation of New Session: {result}')
            dd = await conn.get(session_id)
        except Exception as err:
            logging.exception(err)
            return False
        try:
            session = SessionData(
                db=conn,
                identity=session_id,
                data=data,
                new=True,
                max_age=self.max_age
            )
        except Exception as err:
            print(err)
            logging.exception(f'Error creating Session Data: {err!s}')
        # Saving Session Object:
        session[SESSION_KEY] = session_id
        request[SESSION_OBJECT] = session
        request[SESSION_KEY] = session_id
        request["session"] = session
        return session



"""ADFSAuth.

Description: Backend Authentication/Authorization using Okta Service.
"""
import rapidjson
import logging
import asyncio
import base64
from aiohttp import web, hdrs
from .base import BaseAuthBackend
from typing import List, Dict, Any
import jwt
# needed by ADFS
from xml.etree import ElementTree
import requests
import requests.adapters
from urllib3.util.retry import Retry
from cryptography.hazmat.backends.openssl.backend import backend
from cryptography.x509 import load_der_x509_certificate

from navigator.exceptions import (
    NavException,
    UserDoesntExists,
    InvalidAuth,
    FailedAuth
)
from navigator.conf import (
    ADFS_SERVER,
    ADFS_CLIENT_ID,
    ADFS_RELYING_PARTY_ID,
    ADFS_TENANT_ID,
    ADFS_RESOURCE,
    ADFS_AUDIENCE,
    ADFS_ISSUER,
    USERNAME_CLAIM,
    GROUP_CLAIM,
    ADFS_CLAIM_MAPPING,
    ADFS_LOGIN_REDIRECT_URL,
    AZURE_AD_SERVER
)


def load_certs():
    """Extract token signing certificates."""
    xml_tree = ElementTree.fromstring(response.content)
    cert_nodes = xml_tree.findall(
            "./{urn:oasis:names:tc:SAML:2.0:metadata}RoleDescriptor"
            "[@{http://www.w3.org/2001/XMLSchema-instance}type='fed:SecurityTokenServiceType']"
            "/{urn:oasis:names:tc:SAML:2.0:metadata}KeyDescriptor[@use='signing']"
            "/{http://www.w3.org/2000/09/xmldsig#}KeyInfo"
            "/{http://www.w3.org/2000/09/xmldsig#}X509Data"
            "/{http://www.w3.org/2000/09/xmldsig#}X509Certificate"
    )
    signing_certificates = [node.text for node in cert_nodes]
    new_keys = []
    for cert in signing_certificates:
        logging.debug("Loading public key from certificate: %s", cert)
        cert_obj = load_der_x509_certificate(
            base64.b64decode(cert), backend
        )
        new_keys.append(
            cert_obj.public_key()
        )
    return new_keys

class ADFSAuth(BaseAuthBackend):
    """ADFSAuth.

    Description: Authentication Backend using
    Active Directory Federation Service (ADFS).
    """
    user_attribute: str = "user"
    username_attribute: str = "username"
    pwd_atrribute: str = "password"
    _credentials: Dict = {}
    _adfs: Any = None
    version = 'v1.0'

    def configure(self, app, router, handler):
        # URIs:
        router = app.router
        if ADFS_TENANT_ID:
            self.server = AZURE_AD_SERVER
            self.tenant_id = ADFS_TENANT_ID
            self.username_claim = "upn"
            self.groups_claim = "groups"
            self.claim_mapping = ADFS_CLAIM_MAPPING
        else:
            self.server = ADFS_SERVER
            self.tenant_id = 'adfs'
            self.username_claim = USERNAME_CLAIM
            self.groups_claim = GROUP_CLAIM
            self.claim_mapping = ADFS_CLAIM_MAPPING
        self.base_uri = f"https:://{self.server}/"
        self.end_session_endpoint = f"https://{self.server}/{self.tenant_id}/ls/?wa=wsignout1.0"
        self.issuer = f"https://{self.server}/{self.tenant_id}/services/trust"
        self.authorization_endpoint = f"https://{self.server}/{self.tenant_id}/oauth2/authorize/"
        self.token_endpoint = f"https://{self.server}/{self.tenant_id}/oauth2/token"
        async def _setup_adfs(app):
            pass
        asyncio.get_event_loop().run_until_complete(_setup_adfs(app))
        # creating the Paths
        router.add_route(
            "GET",
            "/auth/adfs/callback",
            self.finish_auth,
            name="adfs_complete_login"
        )
        router.add_route(
            "GET",
            "/auth/adfs/login",
            self.authenticate,
            name="adfs_api_login"
        )
        router.add_route(
            "GET",
            "/auth/adfs/logout",
            self.logout,
            name="adfs_api_logout"
        )
        router.add_route(
            "GET",
            "/auth/adfs/complete_logout",
            self.finish_logout,
            name="adfs_complete_logout"
        )
        # executing parent configurations
        super(ADFSAuth, self).configure(app, router, handler)

    async def get_payload(self, request):
        ctype = request.content_type
        if request.method == "GET":
            try:
                user = request.query.get(self.username_attribute, None)
                password = request.query.get(self.pwd_atrribute, None)
                return [user, password]
            except Exception:
                return None
        elif ctype in ("multipart/mixed", "application/x-www-form-urlencoded"):
            data = await request.post()
            if len(data) > 0:
                user = data.get(self.username_attribute, None)
                password = data.get(self.pwd_atrribute, None)
                return [user, password]
            else:
                return None
        elif ctype == "application/json":
            try:
                data = await request.json()
                user = data[self.username_attribute]
                password = data[self.pwd_atrribute]
                return [user, password]
            except Exception:
                return None
        else:
            return None

    async def authenticate(self, request):
        """ Authenticate, refresh or return the user credentials.

        Description: This function returns the ADFS authorization URL.
        """
        app = request.app
        # print(app)
        absolute_uri = str(request.url)
        print('URL: ', absolute_uri)
        domain_url = absolute_uri.replace(str(request.rel_url), '')
        print('DOMAIN: ', domain_url)
        self.end_authorization_endpoint = f"{domain_url}/auth/adfs/callback"
        try:
            self.redirect_uri = "{}{}".format(
                domain_url,
                app.router["home"].url_for()
            )
        except Exception as err:
            print(err)
            self.redirect_uri = ADFS_LOGIN_REDIRECT_URL
        print('REDIRECT: ', self.redirect_uri)
        try:
            self.state = base64.urlsafe_b64encode(self.redirect_uri.encode()).decode()
            query_params = {
                "client_id": ADFS_CLIENT_ID,
                "response_type": "code",
                "redirect_uri": self.end_authorization_endpoint,
                "resource": ADFS_RESOURCE,
                "response_mode": "query",
                "state": self.state,
                "scope": "openid",
                "grant_type": "client_credentials"
            }
            uri = "{base_uri}?{query_params}".format(
                base_uri=self.authorization_endpoint,
                query_params=requests.compat.urlencode(query_params)
            )
            # Step A: redirect
            logging.debug(f'ADFS URI: {uri}')
            return web.HTTPFound(uri)
        except Exception as err:
            print('HERE: ', err)
            raise NavException(
                f"Client doesn't have info for ADFS Authentication: {err}"
            )

    async def finish_auth(self, request):
        # self.end_authorization_endpoint = "http://navigator.dev.mobileinsight.com/oauth2/callback"
        try:
            response = {key: val for (key, val) in request.query.items()}
            authorization_code = response['code']
            state = response['state']
            request_id = response['client-request-id']
        except Exception as err:
            print(err)
            raise NavException(
                f"ADFS: Invalid Callback response: {err}"
            )
        print(authorization_code, state, request_id)
        logging.debug("Received authorization token: " + code)
        # getting an Access Token
        query_params = {
            "client_id": ADFS_CLIENT_ID,
            "grant_type": "authorization_code",
            "code": authorization_code,
            "redirect_uri": self.end_authorization_endpoint
        }
        query_params = requests.compat.urlencode(query_params)
        print(query_params)
        headers = {
            "Content-Type": "application/x-www-form-urlencoded"
        }
        try:
            exchange = requests.post(
                self.token_endpoint,
                headers=headers,
                data=query_params
            )
            if exchange.status_code == 400:
                logging.error("ADFS server returned an error: " + exchange.json()["error_description"])
                # raise PermissionDenied

            if exchange.status_code != 200:
                logging.error("Unexpected ADFS response: " + exchange.content.decode())
                # raise PermissionDenied
            ## processing the exchange response:
            response = exchange.json()
            access_token = response["access_token"]
            token_type = response["token_type"] # ex: Bearer
            id_token = response["id_token"]
            logging.debug(f"Received access token: {access_token}")
            claims = jwt.decode(
                id_token,
                algorithms=['RS256', 'RS384', 'RS512'],
                verify=False
            )
            logging.debug(f"JWT claims:\n {claims}")
            claims = None
            # validate:
            #
        except Exception as err:
            print(err)
        return web.Response(body="Hello World")

    async def logout(self, request):
        # first: removing the existing session
        # second: redirect to SSO logout
        logging.debug(f'ADFS LOGOUT URI: {self.end_session_endpoint}')
        return web.HTTPFound(self.end_session_endpoint)

    async def finish_logout(self, request):
        pass

    async def check_credentials(self, request):
        """ Authentication and create a session."""
        return True



#!/usr/bin/env python
"""Navigator
    Web Framework based on aiohttp, with batteries included.
See:
https://github.com/phenobarbital/navigator-api/tree/master
"""
from os import path
from setuptools import find_packages, setup


def get_path(filename):
    return path.join(path.dirname(path.abspath(__file__)), filename)


def readme():
    with open(get_path('README.md')) as readme:
        return readme.read()

with open(get_path('navigator/version.py')) as meta:
    exec(meta.read())

setup(
    name=__title__,
    version=__version__,
    python_requires=">=3.8.0",
    url="https://github.com/phenobarbital/navigator-api",
    description=__description__,
    platforms=['POSIX'],
    long_description=readme(),
    long_description_content_type="text/markdown",
    classifiers=[
        "Development Status :: 4 - Beta",
        "Intended Audience :: Developers",
        "Intended Audience :: Information Technology",
        "Intended Audience :: System Administrators",
        "Topic :: Software Development :: Build Tools",
        "Topic :: Software Development :: Libraries :: Application Frameworks",
        "Topic :: Software Development :: Libraries :: Python Modules",
        "Topic :: Software Development :: Libraries",
        "Programming Language :: Python :: 3.8",
    ],
    author=__author__,
    author_email=__author_email__,
    packages=find_packages(exclude=('tests', 'docs', )),
    include_package_data=True,
    license=__license__,
    setup_requires=[
        "wheel==0.37.0",
        "Cython==0.29.28",
        "asyncio==3.4.3",
        "cchardet==2.1.7",
        'cryptography==3.4.7',
        "cpython==0.0.6"
    ],
    install_requires=[
        "wheel==0.37.0",
        "Cython==0.29.28",
        "asyncio==3.4.3",
        "uvloop==0.16.0",
        "asyncdb==2.0.11",
        "navconfig==0.7.4",
        "async-notify==0.5.7",
        "aiofile==3.7.4",
        "aiofiles==0.8.0",
        "sockjs==0.11.0",
        "PySocks==1.7.1",
        "aiodns==3.0.0",
        "asn1crypto==1.4.0",
        "aiohttp-jrpc==0.1.0",
        "PyJWT==2.3.0",
        "pycryptodome==3.14.1",
        "rncryptor==3.3.0",
        "aiohttp-jinja2==1.5",
        "aiohttp-cors==0.7.0",
        "aiohttp-sse==2.1.0",
        "aiosocks==0.2.6",
        "aiohttp-swagger==1.0.16",
        "pydantic==1.9.0",
        "aiohttp-utils==3.1.1",
        "msal==1.17.0",
        "aiogoogle==3.1.2",
        "okta-jwt-verifier==0.2.3",
        "aiologstash==2.0.0",
        "aiohttp-debugtoolbar==0.6.0",
        "jsonpickle==2.1.0",
        'slugify==0.0.1',
        "platformdirs==2.5.1"
    ],
    tests_require=[
            'pytest>=5.4.0',
            'coverage',
            'pytest-asyncio==0.14.0',
            'pytest-xdist==2.1.0',
            'pytest-assume==2.4.2'
    ],
    project_urls={
        'Source': 'https://github.com/phenobarbital/navigator-api',
        'Funding': 'https://paypal.me/phenobarbital',
        'Say Thanks!': 'https://saythanks.io/to/phenobarbital',
    },
)



"""
Base Command for App Creation.
"""
import os
import sys
import json
import asyncio
import logging
from pathlib import Path

from aiofile import AIOFile

from . import BaseCommand, CommandError


logger = logging.getLogger("Navigator.creator")
loop = asyncio.get_event_loop()


def read_file(path, filename):
    f = open(path.joinpath("navigator", "commands", "templates", filename), "r")
    return f.read()


def create_dir(dir, name, touch_init: bool = False):
    try:
        path = dir.joinpath(name)
        path.mkdir(parents=True, exist_ok=True)
        if touch_init is True:
            # create a __init__ file
            save_file(path, "__init__.py", content="#!/usr/bin/env python3")
    except FileExistsError as exc:
        pass


def delete_file(dir, name):
    path = dir.joinpath(name)
    try:
        path.unlink(missing_ok=True)
        return True
    except Exception as err:
        print(err)
        return False


def save_file(dir, filename, content):
    async def main(filename, content):
        try:
            path = dir.joinpath(filename)
            async with AIOFile(path, "w+") as afp:
                await afp.write(content)
                await afp.fsync()
            return True
        except Exception as err:
            print(err)
            logging.error(err)
            return False

    return loop.run_until_complete(main(filename, content))


class AppCommand(BaseCommand):
    help = "Application Creation Commands for Navigator"

    def parse_arguments(self, parser):
        parser.add_argument("--program", type=str)
        parser.add_argument("--is_program", type=bool)
        parser.add_argument("--file_ini", type=str)
        parser.add_argument("--enable_models", type=bool)

    def create(self, options, **kwargs):
        """
        Create can used to create a new Application from scratch
        """
        path = Path(kwargs["project_path"]).resolve()
        try:
            urls = read_file(path, "urls.py.tpl")
            ini = read_file(path, "init.app.tpl")
            views = read_file(path, "views.py.tpl")
        except Exception as err:
            self.write(f":: Error getting templates: {err!s}", level="WARN")
            self.write(f":: Directory: {path!s}", level="WARN")
            output = "Failed."
            return output

        output = "Application Created."
        if options.debug:
            self.write(":: Creating a New Navigator Application", level="INFO")
            self.write("= wait a few seconds", level="WARN")
        # apps, etc, env, services, settings, static/images/js/css, templates
        self.write("* First Step: Creating Directory structure")
        create_dir(path, "apps", touch_init=True)
        app_path = path.joinpath('apps')

        try:
            program = options.program
        except Exception as err:
            self.write(f":: Error getting Program: {err!s}", level="ERROR")
            output = "Failed."
            return output

        if options.is_program:
            ini = read_file(path, "program.app.tpl")

        ini = ini.format(program=program)
        # first: create the program path:
        create_dir(app_path, program)
        # also, the template folder
        program_path = app_path.joinpath(program)
        create_dir(program_path, "templates")
        create_dir(program_path, "images")

        self.write("* Second Step: Creation of Empty Application Structure")
        save_file(program_path, "__init__.py", ini)
        if options.file_ini:
            # create an empty ini file:
            save_file(path, f"{program}.ini")


        # saving urls and views files:
        save_file(program_path, "urls.py", urls)
        save_file(program_path, "views.py", views)

        if options.enable_models:
            models = read_file(path, "models.py.tpl")
            save_file(program_path, "models.py", models)
        return output



#!/usr/bin/env python3
import argparse
import ssl
import sys
import typing
import aiohttp_cors
import logging
import signal
import sockjs
import traceback
from aiohttp import web
from concurrent.futures import ThreadPoolExecutor
from functools import wraps
from typing import (
    List,
    Dict,
    Any,
    Callable,
    Optional
)
# make asyncio use the event loop provided by uvloop
import asyncio
import uvloop
from navigator.conf import (
    DEBUG,
    APP_NAME,
    APP_HOST,
    APP_PORT,
    EMAIL_CONTACT,
    INSTALLED_APPS,
    LOCAL_DEVELOPMENT,
    Context,
    SSL_CERT,
    SSL_KEY,
    loglevel,
    TEMPLATE_DIR,
    CACHE_URL,
    default_dsn
)
from navigator.connections import PostgresPool, RedisPool
from navigator.applications import AppBase, AppHandler, app_startup
# Exception Handlers
from navigator.handlers import (
    nav_exception_handler,
    shutdown
)
from navigator.templating import TemplateParser
# websocket resources
from navigator.resources import WebSocket, channel_handler


asyncio.set_event_loop_policy(uvloop.EventLoopPolicy())


def Response(
    content: Any = None,
    text: str = "",
    body: Any = None,
    status: int = 200,
    headers: dict = None,
    content_type: str = "text/plain",
    charset: str = "utf-8",
) -> web.Response:
    """
    Response.
    Web Response Definition for Navigator
    """
    response = {
        "content_type": content_type,
        "charset": charset,
        "status": status
    }
    if headers:
        response["headers"] = headers
    if isinstance(content, str) or text is not None:
        response["text"] = content if content else text
    else:
        response["body"] = content if content else body
    return web.Response(**response)


class Application(object):
    def __init__(
        self,
        app: AppHandler = None,
        enable_swagger: bool = False,
        enable_debugtoolbar: bool = False,
        enable_jinja_parser: bool = True,
        use_ssl: bool = False,
        title: str = '',
        description: str = 'NAVIGATOR APP',
        contact: str = '',
        version: str = "0.0.1",
        swagger_options: Dict = {},

        *args,
        **kwargs
    ) -> None:
        self.version = version
        self.enable_debugtoolbar = enable_debugtoolbar
        self.enable_swagger = enable_swagger
        self.use_ssl = use_ssl
        self.description = description
        self.contact = contact
        if not contact:
            self.contact = EMAIL_CONTACT
        self.title = title
        if not title:
            self.title = APP_NAME
        self.swagger_options = swagger_options
        self.enable_jinja_parser = enable_jinja_parser
        # configuring asyncio loop
        # TODO: work in an exception handler for NAV
        self._loop = asyncio.get_event_loop()
        self._loop.set_exception_handler(nav_exception_handler)
        # May want to catch other signals too
        signals = (signal.SIGHUP, signal.SIGTERM, signal.SIGINT)
        for s in signals:
            self._loop.add_signal_handler(
                s, lambda s=s: asyncio.create_task(
                    shutdown(self._loop, s)
                )
            )
        self.parser = argparse.ArgumentParser(
            description="Navigator App"
        )
        self.parser.add_argument("--path")
        self.parser.add_argument("--host")
        self.parser.add_argument("--port")
        self.parser.add_argument(
            "-d", "--debug", action="store_true", help="Enable Debug"
        )
        self.parser.add_argument(
            "--traceback", action="store_true", help="Return Traceback on Error"
        )
        self.parse_arguments()
        if not app:
            # create an instance of AppHandler
            self.app = AppBase(Context)
        else:
            self.app = app(Context)
        # getting the application Logger
        self._logger = self.get_logger(self.app.Name)

    def parse_arguments(self):
        args = self.parser.parse_args()
        try:
            self.path = args.path
        except (KeyError, ValueError, TypeError):
            self.path = None
        try:
            self.host = args.host
            if not self.host:
                self.host = APP_HOST
        except (KeyError, ValueError, TypeError):
            self.host = APP_HOST
        try:
            self.port = args.port
            if not self.port:
                self.port = APP_PORT
        except (KeyError, ValueError, TypeError):
            self.port = APP_PORT
        try:
            self.debug = args.debug
        except (KeyError, ValueError, TypeError):
            self.debug = DEBUG

    def get_app(self) -> web.Application:
        return self.app.App

    def __setitem__(self, k, v):
        self.app.App[k] = v

    def __getitem__(self, k):
        return self.app.App[k]

    def get_logger(self, name:str = "Navigator"):
        logging_format = f"[%(asctime)s] %(levelname)-5s %(name)-{len(name)}s "
        logging_format += "%(message)s"
        logging.basicConfig(
            format=logging_format, level=loglevel, datefmt="%Y:%m:%d %H:%M:%S"
        )
        return logging.getLogger(name)

    def setup_app(self) -> web.Application:
        app = self.get_app()
        if self.enable_jinja_parser is True:
            try:
                parser = TemplateParser(
                    directory=TEMPLATE_DIR
                )
                app['template'] = parser
            except Exception:
                raise
        # create the pool-based connections (shared):
        name = app["name"]
        redis = RedisPool(
            dsn=CACHE_URL,
            name=f"NAV-{name!s}",
            loop=self._loop
        )
        redis.configure(app)
        app['redis'] = redis.connection()
        # Database Pool:
        db = PostgresPool(
            dsn=default_dsn,
            name=f"NAV-{name!s}",
            loop=self._loop
        )
        db.configure(app)
        app['database'] = db.connection()
        # setup The Application and Sub-Applications Startup
        app_startup(INSTALLED_APPS, app, Context)
        # Configure Routes
        self.app.configure()
        cors = aiohttp_cors.setup(
            app,
            defaults={
                "*": aiohttp_cors.ResourceOptions(
                    allow_credentials=True,
                    expose_headers="*",
                    allow_methods="*",
                    allow_headers="*",
                    max_age=3600,
                )
            },
        )
        self.app.setup_cors(cors)
        self.app.setup_docs()
        return app

    def add_websockets(self) -> None:
        """
        add_websockets.
        description: enable support for websockets in the main App
        """
        app = self.get_app()
        if self.debug:
            logging.debug("Enabling WebSockets")
        # websockets
        app.router.add_route("GET", "/ws", WebSocket)
        # websocket channels
        app.router.add_route("GET", "/ws/{channel}", channel_handler)

    def add_routes(self, routes: list) -> None:
        """
        add_route
        description: append a route to routes dict
        """
        # TODO: avoid to add same route different times
        # self._routes.append(route)
        self.get_app().add_routes(routes)

    def add_static(self, route: str, path: str):
        """
        add_static
        description: register new route to static path.
        """
        self.get_app().add_static(route, path)

    def add_view(self, route: str, handler: Any):
        self.get_app().router.add_view(route, handler)

    def threaded_func(self, func: Callable, threaded: bool = False):
        @wraps(func)
        async def _wrap(request):
            result = None
            try:
                if threaded:

                    def blocking_function():
                        return asyncio.new_event_loop().run_until_complete(
                            func(request)
                        )

                    result = await self._loop.run_in_executor(
                        ThreadPoolExecutor(max_workers=1), blocking_function
                    )
                else:
                    result = await func(request)
                return result
            except Exception as err:
                self._logger.exception(err)
        return _wrap

    def route(self, route: str, method: str = "GET", threaded: bool = False):
        """
        route.
        description: decorator for register a new HTTP route.
        """

        def _decorator(func):
            self.app.App.router.add_route(
                method, route, self.threaded_func(func, threaded)
            )
            return func

        return _decorator

    def add_get(self, route: str, threaded: bool = False) -> Callable:
        def _decorator(func):
            self.app.App.router.add_get(
                route, self.threaded_func(func, threaded), allow_head=False
            )
            return func

        return _decorator

    def Response(self, content: Any) -> web.Response:
        return web.Response(text=content)

    def get(self, route: str):
        def _decorator(func):
            self.app.App.router.add_get(route, func)

            @wraps(func)
            async def _wrap(request, *args, **kwargs):
                try:
                    return f"{func(request, args, **kwargs)}"
                except Exception as err:
                    self._logger.exception(err)

            return _wrap

        return _decorator

    @property
    def router(self):
        return self.app.App.router

    def post(self, route: str):
        def _decorator(func):
            self.app.App.router.add_post(route, func)

            @wraps(func)
            async def _wrap(request, *args, **kwargs):
                try:
                    return f"{func(request, args, **kwargs)}"
                except Exception as err:
                    self._logger.exception(err)

            return _wrap

        return _decorator

    def add_sock_endpoint(
        self, handler: Callable, name: str, route: str = "/sockjs/"
    ) -> None:
        app = self.get_app()
        sockjs.add_endpoint(app, handler, name=name, prefix=route)

    def run(self):
        # getting the resource App
        app = self.setup_app()
        if self.debug:
            logging.debug('Running in DEBUG mode.')
        if self.enable_swagger is True:
            # previous to run, setup swagger:
            # auto-configure swagger
            from aiohttp_swagger import setup_swagger
            setup_swagger(
                app,
                api_base_url="/",
                title=self.title,
                api_version=self.version,
                description=self.description,
                swagger_url="/api/v1/doc",
                ui_version=3,
                **self.swagger_options
            )
        if self.debug is True:
            if LOCAL_DEVELOPMENT:
                if self.enable_debugtoolbar is True:
                    import aiohttp_debugtoolbar
                    from aiohttp_debugtoolbar import toolbar_middleware_factory
                    aiohttp_debugtoolbar.setup(
                        app,
                        hosts=[self.host, "127.0.0.1", "::1"],
                        enabled=True,
                        path_prefix="/_debug",
                    )
            try:
                web.run_app(app, host=self.host, port=self.port)
            except Exception as err:
                print(traceback.format_exc())
                print(err)
        else:
            if self.use_ssl:
                ssl_context = ssl.SSLContext(ssl.PROTOCOL_TLSv1_2)
                ssl_context.load_cert_chain(SSL_CERT, SSL_KEY)
                web.run_app(
                    app, host=self.host, port=self.port, ssl_context=ssl_context
                )
            if self.path:
                web.run_app(app, path=self.path)
            else:
                web.run_app(app, host=self.host, port=self.port)



"""Base Class for all Session Storages."""

import abc
import json
import uuid
import time
import asyncio
import logging
from aiohttp import web
from aiohttp.web_middlewares import Handler, middleware
from typing import (
    Any,
    Callable,
    Dict,
    List,
    Iterator,
    Mapping,
    MutableMapping,
    Optional
)
from navigator.conf import (
    SESSION_NAME,
    SESSION_PREFIX,
    SESSION_TIMEOUT,
    SESSION_KEY,
    SESSION_OBJECT,
    SESSION_STORAGE
)
import jsonpickle


class SessionData(MutableMapping[str, Any]):
    """Session dict-like object.

    TODO: Support saving directly into Storage.
    """

    _data: Dict[str, Any] = {}
    _db: Callable = None

    def __init__(
        self,
        db: Callable, *,
        data: Optional[Mapping[str, Any]] = {},
        new: bool = False,
        identity: Optional[Any] = None,
        max_age: Optional[int] = None
    ) -> None:
        self._changed = False
        self._data = {}
        self._db = db
        self._identity = data.get(SESSION_KEY, None) if data else identity
        if not self._identity:
            self._identity = uuid.uuid4().hex
        self._new = new if data != {} else True
        self._max_age = max_age if max_age else None
        created = data.get('created', None) if data else None
        now = int(time.time())
        age = now - created if created else now
        if max_age is not None and age > max_age:
            session_data = None
        if self._new or created is None:
            self._created = now
        else:
            self._created = created

        if data is not None:
            self._data.update(data)

    def __repr__(self) -> str:
        return '<{} [new:{}, created:{}] {!r}>'.format(
            'NAV-Session ', self.new, self.created, self._data
        )

    @property
    def new(self) -> bool:
        return self._new

    @property
    def identity(self) -> Optional[Any]:  # type: ignore[misc]
        return self._identity

    @property
    def created(self) -> int:
        return self._created

    @property
    def empty(self) -> bool:
        return not bool(self._data)

    @property
    def max_age(self) -> Optional[int]:
        return self._max_age

    @max_age.setter
    def max_age(self, value: Optional[int]) -> None:
        self._max_age = value

    @property
    def is_changed(self) -> bool:
        return self._changed

    def changed(self) -> None:
        self._changed = True

    def session_data(self) -> Dict:
        return self._data

    def invalidate(self) -> None:
        self._changed = True
        self._data = {}

    # Magic Methods
    def __len__(self) -> int:
        return len(self._data)

    def __iter__(self) -> Iterator[str]:
        return iter(self._data)

    def __contains__(self, key: object) -> bool:
        return key in self._data

    def __getitem__(self, key: str) -> Any:
        return self._data[key]

    def __setitem__(self, key: str, value: Any) -> None:
        self._data[key] = value
        self._changed = True
        # TODO: also, saved into redis automatically

    def __delitem__(self, key: str) -> None:
        del self._data[key]
        self._changed = True

    def __getattr__(self, key: str) -> Any:
        return self._data[key]
    
    def encode(self, obj: Any) -> str:
        """encode

            Encode an object using jsonpickle.
        Args:
            obj (Any): Object to be encoded using jsonpickle

        Raises:
            RuntimeError: Error converting data to json.

        Returns:
            str: json version of the data
        """
        try:
            return jsonpickle.encode(obj)
        except Exception as err:
            raise RuntimeError(err)
    
    def decode(self, key: str) -> Any:
        """decode.

            Decoding a Session Key using jsonpickle.
        Args:
            key (str): key name.

        Raises:
            RuntimeError: Error converting data from json.

        Returns:
            Any: object converted.
        """
        try:
            value = self._data[key]
            return jsonpickle.decode(value)
        except Exception as err:
            raise RuntimeError(err)


class AbstractStorage(metaclass=abc.ABCMeta):

    def id_factory(self) -> str:
        return uuid.uuid4().hex

    def __init__(self, max_age: int = None, secure: bool = None) -> None:
        if not max_age:
            self.max_age = SESSION_TIMEOUT
        else:
            self.max_age = max_age

    def configure_session(self, app: web.Application) -> None:
        """Configure the Middleware for NAV Session."""
        app.middlewares.append(
            session_middleware(app, self)
        )

    @abc.abstractmethod
    async def new_session(
        self,
        request: web.Request,
        data: Dict = None
    ) -> SessionData:
        pass

    @abc.abstractmethod
    async def load_session(
        self,
        request: web.Request,
        new: bool = False
    ) -> SessionData:
        pass

    @abc.abstractmethod
    async def get_session(self, request: web.Request) -> SessionData:
        pass

    @abc.abstractmethod
    async def save_session(self,
        request: web.Request,
        response: web.StreamResponse,
        session: SessionData
    ) -> None:
        pass

    @abc.abstractmethod
    async def invalidate(
        self,
        request: web.Request,
        session: SessionData
    ) -> None:
        """Try to Invalidate the Session in the Storage."""
        pass

    async def forgot(self, request):
        """Delete a User Session."""
        session = await self.get_session(request)
        await self.invalidate(request, session)
        request["session"] = None
        try:
            del request[SESSION_KEY]
            del request[SESSION_OBJECT]
        except Exception as err:
            print('Error: ', err)
        finally:
            return True

"""
 Basic Middleware for Session System
"""
def session_middleware(
        app: web.Application,
        storage: 'AbstractStorage'
) -> middleware:
    """Middleware to attach Session Storage to every Request."""
    if not isinstance(storage, AbstractStorage):
        raise RuntimeError(f"Expected an AbstractStorage got {storage!s}")

    @web.middleware
    async def middleware(
            request: web.Request,
            handler: Handler
    ) -> web.StreamResponse:
        request[SESSION_STORAGE] = storage
        try:
            response = await handler(request)
        except web.HTTPException as exc:
            raise exc
        if not isinstance(response, (web.StreamResponse, web.HTTPException)):
            # likely got websocket or streaming
            return response
        if response.prepared:
            raise RuntimeError(
                "We Cannot save session data into on prepared responses"
            )
        session = request.get(SESSION_OBJECT)
        if isinstance(session, SessionData):
            if session.is_changed:
                await storage.save_session(request, response, session)
        return response

    return middleware



""" Authorization based on HOSTS lists."""

from .base import BaseAuthzHandler
from navigator.conf import HOSTS
from aiohttp import web
import logging


class authz_hosts(BaseAuthzHandler):
    """
    BasicHosts.
       Use for basic Host authorization, simply creating a list of allowed hosts
    """

    async def check_authorization(self, request: web.Request) -> bool:
        if request.host in HOSTS:
            logging.debug('Authorized based on HOST Authorization')
            return True
        try:
            if request.headers["origin"] in HOSTS:
                logging.debug('Authorization based on HOST')
                return True
        except KeyError:
            return False
        return False



"""Navigator Meta information."""

__title__ = 'navigator-api'
__description__ = ('Navigator Web Framework based on aiohttp, '
                   'with batteries included.')
__version__ = '2.0.8'
__author__ = 'Jesus Lara'
__author_email__ = 'jesuslarag@gmail.com'
__license__ = 'BSD'
